@@ -37,6 +37,8 @@ setMethod(f = "draw_heatmap_body",
     raster_device = object@heatmap_param$raster_device
     raster_quality = object@heatmap_param$raster_quality
     raster_device_param = object@heatmap_param$raster_device_param
+    raster_by_magick = object@heatmap_param$raster_by_magick
+    raster_magick_filter = object@heatmap_param$raster_magick_filter
     if(length(raster_device_param) == 0) raster_device_param = list()
 
     pushViewport(viewport(name = paste(object@name, "heatmap_body", kr, kc, sep = "_"), ...))
@@ -65,7 +67,8 @@ setMethod(f = "draw_heatmap_body",
             tiff = c("grDevices", "tiff", "readTIFF"),
             CairoPNG = c("Cairo", "png", "readPNG"),
             CairoJPEG = c("Cairo", "jpeg", "readJPEG"),
-            CairoTIFF = c("Cairo", "tiff", "readTIFF")
+            CairoTIFF = c("Cairo", "tiff", "readTIFF"),
+            agg_png = c("ragg", "png", "readPNG")
         )
         if(!requireNamespace(device_info[1])) {
             stop_wrap(paste0("Need ", device_info[1], " package to write image."))
@@ -73,37 +76,76 @@ setMethod(f = "draw_heatmap_body",
         if(!requireNamespace(device_info[2])) {
             stop_wrap(paste0("Need ", device_info[2], " package to read image."))
         }
-        # can we get the size of the heatmap body?
-        heatmap_width = convertWidth(unit(1, "npc"), "bigpts", valueOnly = TRUE)
-        heatmap_height = convertHeight(unit(1, "npc"), "bigpts", valueOnly = TRUE)
-        if(heatmap_width <= 0 || heatmap_height <= 0) {
-            stop_wrap("The width or height of the raster image is zero, maybe you forget to turn off the previous graphic device or it was corrupted. Run `dev.off()` to close it.")
+
+        if(raster_device %in% c("png", "jpeg", "tiff")) {
+            if(! "type" %in% names(raster_device_param)) {
+                if(capabilities("cairo")) {
+                    raster_device_param$type = "cairo"
+                }
+            }
         }
 
+        # can we get the size of the heatmap body?
+        heatmap_width_pt = max(1, ceiling(convertWidth(unit(1, "npc"), "bigpts", valueOnly = TRUE)))
+        heatmap_height_pt = max(1, ceiling(convertHeight(unit(1, "npc"), "bigpts", valueOnly = TRUE)))
+
         matrix_is_resized = FALSE
-        if(object@heatmap_param$raster_resize) {
-            if(heatmap_width < nc && heatmap_height < nr) {
-                mat2 = resize_matrix(mat, nr = heatmap_height, nc = heatmap_width)
+        # resize on the matrix
+        raster_resize_mat = object@heatmap_param$raster_resize_mat
+        if(!identical(raster_resize_mat, FALSE)) {
+            if(is.logical(raster_resize_mat)) {
+                raster_resize_mat_fun = function(x) mean(x, na.rm = TRUE)
+            } else {
+                if(!inherits(raster_resize_mat, "function")) {
+                    stop_wrap("`raster_resize_mat` should be set as logical scalar or a function.")
+                }
+                raster_resize_mat_fun = raster_resize_mat
+            }
+
+            if(heatmap_width_pt < nc && heatmap_height_pt < nr) {
+                mat2 = resize_matrix(mat, nr = heatmap_height_pt, nc = heatmap_width_pt, fun = raster_resize_mat_fun)
                 matrix_is_resized = TRUE
-            } else if(heatmap_width < nc) {
-                mat2 = resize_matrix(mat, nr = nr, nc = heatmap_width)
+            } else if(heatmap_width_pt < nc) {
+                mat2 = resize_matrix(mat, nr = nr, nc = heatmap_width_pt, fun = raster_resize_mat_fun)
                 matrix_is_resized = TRUE
-            } else if(heatmap_height < nr) {
-                mat2 = resize_matrix(mat, nr = heatmap_height, nc = nc)
+            } else if(heatmap_height_pt < nr) {
+                mat2 = resize_matrix(mat, nr = heatmap_height_pt, nc = nc, fun = raster_resize_mat_fun)
                 matrix_is_resized = TRUE
             }
         }
 
+        if(matrix_is_resized) {
+            raster_by_magick = FALSE
+        }
+
         temp_dir = tempdir()
-                # dir.create(tmp_dir, showWarnings = FALSE)
-        temp_image = tempfile(pattern = paste0(".heatmap_body_", object@name, "_", kr, "_", kc), tmpdir = temp_dir, fileext = paste0(".", device_info[2]))
-        #getFromNamespace(raster_device, ns = device_info[1])(temp_image, width = heatmap_width*raster_quality, height = heatmap_height*raster_quality)
+        temp_image = tempfile(pattern = paste0(".heatmap_body_", digest::digest(object@name), "_", kr, "_", kc), tmpdir = temp_dir, fileext = paste0(".", device_info[2]))
         device_fun = getFromNamespace(raster_device, ns = device_info[1])
 
-        do.call(device_fun, c(list(filename = temp_image, width = max(c(heatmap_width*raster_quality, 1)), height = max(c(heatmap_height*raster_quality, 1))), raster_device_param))
+        if(raster_by_magick) {
+            temp_image_width = ceiling(max(heatmap_width_pt, nc, 1))
+            temp_image_height = ceiling(max(heatmap_height_pt, nr, 1))
+        } else if(matrix_is_resized) {
+            temp_image_width = ceiling(max(heatmap_width_pt, 1))
+            temp_image_height = ceiling(max(heatmap_height_pt, 1))
+        } else {
+            if(is.character(raster_quality)) {
+                temp_image_width = ceiling(max(heatmap_width_pt, nc, 1))
+                temp_image_height = ceiling(max(heatmap_height_pt, nr, 1))
+            } else {
+                if(raster_quality < 1) raster_quality = 1
+                temp_image_width = ceiling(max(heatmap_width_pt*raster_quality, 1))
+                temp_image_height = ceiling(max(heatmap_height_pt*raster_quality, 1))
+            }
+        }
+        do.call(device_fun, c(list(filename = temp_image, 
+            width = temp_image_width, height = temp_image_height), raster_device_param))
+        if(object@heatmap_param$verbose) {
+            qqcat("saving into a temp image (.@{device_info[2]}) with size @{temp_image_width}x@{temp_image_height}px.\n")
+        }
         if(matrix_is_resized) {
             if(object@heatmap_param$verbose) {
-                qqcat("resize the matrix from (@{nrow(mat)} x @{ncol(mat)}) to (@{nrow(mat2)} x @{ncol(mat2)}).\n")
+                qqcat("resize the matrix from (@{nrow(mat)}x@{ncol(mat)}) to (@{nrow(mat2)}x@{ncol(mat2)}).\n")
             }
             col_matrix2 = map_to_colors(object@matrix_color_mapping, mat2)
             nc2 = ncol(mat2)
@@ -118,62 +160,60 @@ setMethod(f = "draw_heatmap_body",
         if(is.function(layer_fun)) {
             if(length(as.list(formals(layer_fun))) == 7) {
                 layer_fun(column_order[ expand_index[[2]] ], row_order[ expand_index[[1]] ], 
-                    x[expand_index[[2]]], y[expand_index[[1]]],
+                    unit(x[expand_index[[2]]], "npc"), unit(y[expand_index[[1]]], "npc"),
                     unit(rep(1/nc, nrow(expand_index)), "npc"), unit(rep(1/nr, nrow(expand_index)), "npc"),
                     as.vector(col_matrix))
             } else {
                 layer_fun(column_order[ expand_index[[2]] ], row_order[ expand_index[[1]] ], 
-                    x[expand_index[[2]]], y[expand_index[[1]]],
+                    unit(x[expand_index[[2]]], "npc"), unit(y[expand_index[[1]]], "npc"),
                     unit(rep(1/nc, nrow(expand_index)), "npc"), unit(rep(1/nr, nrow(expand_index)), "npc"),
                     as.vector(col_matrix), kr, kc)
             }
         }
         dev.off2()
-        
-        # ############################################
-        # ## make the heatmap body in a another process
-        # temp_R_data = tempfile(pattern = paste0(".heatmap_body_", object@name, "_", kr, "_", kc), tmpdir = temp_dir, fileext = paste0(".RData"))
-        # temp_R_file = tempfile(pattern = paste0(".heatmap_body_", object@name, "_", kr, "_", kc), tmpdir = temp_dir, fileext = paste0(".R"))
-        # if(Sys.info()["sysname"] == "Windows") {
-        #     temp_image = gsub("\\\\", "/", temp_image)
-        #     temp_R_data = gsub("\\\\", "/", temp_R_data)
-        #     temp_R_file = gsub("\\\\", "/", temp_R_file)
-        # }
-        # save(device_fun, device_info, temp_image, heatmap_width, raster_quality, heatmap_height, raster_device_param,
-        #     gp, x, expand_index, nc, nr, col_matrix, row_order, column_order, y,
-        #     file = temp_R_data)
-        # R_cmd = qq("
-        # library(@{device_info[1]})
-        # library(grid)
-        # load('@{temp_R_data}')
-        # do.call('device_fun', c(list(filename = temp_image, width = max(c(heatmap_width*raster_quality, 1)), height = max(c(heatmap_height*raster_quality, 1))), raster_device_param))
-        # grid.rect(x[expand_index[[2]]], y[expand_index[[1]]], width = unit(1/nc, 'npc'), height = unit(1/nr, 'npc'), gp = do.call('gpar', c(list(fill = col_matrix), gp)))
-        # dev.off()
-        # q(save = 'no')
-        # ", code.pattern = "@\\{CODE\\}")
-        # writeLines(R_cmd, con = temp_R_file)
-        # if(grepl(" ", temp_R_file)) {
-        #     if(is_windows()) {
-        #         oe = try(system(qq("\"@{normalizePath(R_binary(), winslash='/')}\" --vanilla < \'@{temp_R_file}\'", code.pattern = "@\\{CODE\\}"), ignore.stdout = TRUE, ignore.stderr = TRUE, show.output.on.console = FALSE), silent = TRUE)
-        #     } else {
-        #         oe = try(system(qq("\"@{normalizePath(R_binary(), winslash='/')}\" --vanilla < \'@{temp_R_file}\'", code.pattern = "@\\{CODE\\}"), ignore.stdout = TRUE, ignore.stderr = TRUE), silent = TRUE)
-        #     }
-        # } else {
-        #     if(is_windows()) {
-        #         oe = try(system(qq("\"@{normalizePath(R_binary(), winslash='/')}\" --vanilla < @{temp_R_file}", code.pattern = "@\\{CODE\\}"), ignore.stdout = TRUE, ignore.stderr = TRUE, show.output.on.console = FALSE), silent = TRUE)
-        #     } else {
-        #         oe = try(system(qq("\"@{normalizePath(R_binary(), winslash='/')}\" --vanilla < @{temp_R_file}", code.pattern = "@\\{CODE\\}"), ignore.stdout = TRUE, ignore.stderr = TRUE), silent = TRUE)
-        #     }
-        # }
-        # ############################################
-        # file.remove(temp_R_data)
-        # file.remove(temp_R_file)
-        # if(inherits(oe, "try-error")) {
-        #     stop(oe)
-        # }
-        image = getFromNamespace(device_info[3], ns = device_info[2])(temp_image)
-        image = as.raster(image)
-        grid.raster(image, width = unit(1, "npc"), height = unit(1, "npc"))
+
+        if(object@heatmap_param$verbose) {
+            qqcat("resize the temp image to a size @{heatmap_width_pt}x@{heatmap_height_pt}px.\n")
+        }
+        if(raster_by_magick) {
+            if(object@heatmap_param$verbose) {
+                qqcat("image is read by magick.\n")
+            }
+            if(!requireNamespace("magick")) {
+                stop_wrap("'magick' package should be installed.")
+            }
+            image = magick::image_read(temp_image)
+            image = magick::image_resize(image, paste0(heatmap_width_pt, "x", heatmap_height_pt, "!"), filter = raster_magick_filter)
+            image = as.raster(image)
+        } else {
+            if(object@heatmap_param$verbose) {
+                qqcat("image is read by @{device_info[2]}::@{device_info[3]}\n")
+            }
+            image = getFromNamespace(device_info[3], ns = device_info[2])(temp_image)
+        }
+        # validate image, there might be white horizontal lines and vertical lines 
+        # image = validate_raster_matrix(image, mat, object@matrix_color_mapping)
+
+        grid.raster(image, width = unit(1, "npc"), height = unit(1, "npc"), interpolate = FALSE)
+
+        ### only for testing the temp image size ###
+        if(ht_opt("__export_image_size__")) {
+            if(inherits(image, "magick-image")) {
+                image = as.raster(image)
+            } else {
+                tf = tempfile()
+                png(tf, width = heatmap_width_pt, height = heatmap_height_pt)
+                grid.raster(image, width = unit(1, "npc"), height = unit(1, "npc"))
+                dev.off()
+                image = as.raster(png::readPNG(tf))
+                file.remove(tf)   
+            }
+            attr(image, "width") = heatmap_width_pt
+            attr(image, "height") = heatmap_height_pt
+            assign(".image", image, envir = .GlobalEnv)
+        }
+        ########################
+
         file.remove(temp_image)
 
     } else {
@@ -199,26 +239,64 @@ setMethod(f = "draw_heatmap_body",
         if(is.function(layer_fun)) {
             if(length(as.list(formals(layer_fun))) == 7) {
                 layer_fun(column_order[ expand_index[[2]] ], row_order[ expand_index[[1]] ], 
-                    x[expand_index[[2]]], y[expand_index[[1]]],
+                    unit(x[expand_index[[2]]], "npc"), unit(y[expand_index[[1]]], "npc"),
                     unit(rep(1/nc, nrow(expand_index)), "npc"), unit(rep(1/nr, nrow(expand_index)), "npc"),
                     as.vector(col_matrix))
             } else {
                 layer_fun(column_order[ expand_index[[2]] ], row_order[ expand_index[[1]] ], 
-                x[expand_index[[2]]], y[expand_index[[1]]],
-                unit(rep(1/nc, nrow(expand_index)), "npc"), unit(rep(1/nr, nrow(expand_index)), "npc"),
-                as.vector(col_matrix), kr, kc)
+                    unit(x[expand_index[[2]]], "npc"), unit(y[expand_index[[1]]], "npc"),
+                    unit(rep(1/nc, nrow(expand_index)), "npc"), unit(rep(1/nr, nrow(expand_index)), "npc"),
+                    as.vector(col_matrix), kr, kc)
             }
         }
     }
 
-    if(!identical(border, FALSE)) {
-        grid.rect(gp = gpar(fill = "transparent", col = border))
+    if(!(identical(border, FALSE) || identical(border, NA))) {
+        border_gp = object@matrix_param$border_gp
+        if(!identical(border, TRUE)) {
+            border_gp$col = border
+        }
+        if("fill" %in% names(border_gp)) {
+            message_wrap("`fill` is ignored in `border_gp`. The value for `fill` is always 'transparent'.")
+        }
+        border_gp$fill = "transparent"
+        grid.rect(gp = border_gp)
     }
 
     upViewport()
 
 })
 
+# check white lines in the RGB matrix, and re-fill the color from mat with color_mapping
+# careful: row orders of rgb and mat are reversed
+# rgb: values change from 0 to 1
+validate_raster_matrix = function(rgb, mat, col_mapping) {
+    if(is.character(rgb)) {
+        dim = dim(rgb)
+        x = col2rgb(rgb)/255
+        rgb = array(dim = c(dim, 3))
+        rgb[, , 1] = matrix(x[1, ], nrow = dim[1], ncol = dim[2], byrow = TRUE)
+        rgb[, , 2] = matrix(x[2, ], nrow = dim[1], ncol = dim[2], byrow = TRUE)
+        rgb[, , 3] = matrix(x[3, ], nrow = dim[1], ncol = dim[2], byrow = TRUE)
+    }
+    # check rows
+    white_row = NULL
+    white_column = NULL
+    if(any( abs(rgb[, 1, 1] - 1) < 1e-10 & abs(rgb[, 1, 2] - 1) < 1e-10 & abs(rgb[, 1, 3] - 1)< 1e-10 )) {
+        white_row = which( abs(rowMeans(rgb[, , 1]) - 1) < 1e-10 & 
+                           abs(rowMeans(rgb[, , 2]) - 1) < 1e-10 & 
+                           abs(rowMeans(rgb[, , 3]) - 1) < 1e-10 )
+    }
+    if(any( abs(rgb[1, , 1] - 1) < 1e-10 & abs(rgb[1, , 2] - 1) < 1e-10 & abs(rgb[1, , 3] - 1)< 1e-10 )) {
+        white_column = which( abs(colMeans(rgb[, , 1]) - 1) < 0.1 & 
+                              abs(colMeans(rgb[, , 2]) - 1) < 0.1 & 
+                              abs(colMeans(rgb[, , 3]) - 1) < 0.1 )
+    }
+
+    # estimate while rows and columns
+    rgb
+}
+
 is_windows = function() {
     tolower(.Platform$OS.type) == "windows"
 }
@@ -427,6 +505,14 @@ setMethod(f = "draw_title",
         "row" = object@row_title_param$just,
         "column" = object@column_title_param$just)
 
+    if(!is.null(ht_opt$TITLE_PADDING)) {
+        title_padding = ht_opt$TITLE_PADDING
+    } else {
+        title_padding = unit(c(0, 0), "points")
+        title_padding[1] = title_padding[1] + unit(5.5, "points") + 
+            convertHeight(grobDescent(textGrob(label = "jA", gp = gp)), "inches")
+    }
+
     if(which == "row") {
         
         pushViewport(viewport(name = paste(object@name, "row_title", k, sep = "_"), clip = FALSE, ...))
@@ -434,9 +520,9 @@ setMethod(f = "draw_title",
         if("border" %in% names(gp2)) gp2$col = gp2$border
         if(any(c("border", "fill") %in% names(gp2))) grid.rect(gp = gp2)
         if(side == "left") {
-            grid.text(title, x = unit(1, "npc") - ht_opt$TITLE_PADDING, rot = rot, just = just, gp = gp)
+            grid.text(title, x = unit(1, "npc") - title_padding[1], rot = rot, just = just, gp = gp)
         } else {
-            grid.text(title, x = ht_opt$TITLE_PADDING, rot = rot, just = just, gp = gp)
+            grid.text(title, x = title_padding[1], rot = rot, just = just, gp = gp)
         }
         upViewport()
     } else {
@@ -445,9 +531,9 @@ setMethod(f = "draw_title",
         if("border" %in% names(gp2)) gp2$col = gp2$border
         if(any(c("border", "fill") %in% names(gp2))) grid.rect(gp = gp2)
         if(side == "top") {
-            grid.text(title, y = ht_opt$TITLE_PADDING, rot = rot, just = just, gp = gp)
+            grid.text(title, y = title_padding[1], rot = rot, just = just, gp = gp)
         } else {
-            grid.text(title, y = unit(1, "npc") - ht_opt$TITLE_PADDING, rot = rot, just = just, gp = gp)
+            grid.text(title, y = unit(1, "npc") - title_padding[1], rot = rot, just = just, gp = gp)
         }
         upViewport()
     }
@@ -527,7 +613,11 @@ setMethod(f = "draw_annotation",
                 anno_mark_param$.pos = .pos
                 anno_mark_param$index = unlist(ro_lt)
                 
-                anno_mark_param$vp_height = convertHeight(unit(1, "npc"), "cm")
+                if(abs(par("din")[2] - grDevices::dev.size("in")[2]) < 1e-6) {
+                    anno_mark_param$vp_height = convertHeight(unit(1, "npc"), "cm")  
+                } else {
+                    anno_mark_param$vp_height = convertHeight(unit(1, "npc"), "cm")*(par("din")[2]/grDevices::dev.size("in")[2])*0.9742
+                }
                 anno_mark_param$vp_width = unit(1, "npc")
                 anno_mark_param$vp_just = "top"
                 anno_mark_param$vp_x = unit(0.5, "npc")
@@ -558,7 +648,11 @@ setMethod(f = "draw_annotation",
                 anno_mark_param$index = unlist(co_lt)
                 
                 anno_mark_param$vp_height = unit(1, "npc")
-                anno_mark_param$vp_width = convertWidth(unit(1, "npc"), "cm")
+                if(abs(par("din")[1] - grDevices::dev.size("in")[1]) < 1e-6) {
+                    anno_mark_param$vp_width = convertWidth(unit(1, "npc"), "cm")
+                } else {
+                    anno_mark_param$vp_width = convertWidth(unit(1, "npc"), "cm")*(par("din")[1]/grDevices::dev.size("in")[1])*0.945
+                }
                 anno_mark_param$vp_just = "left"
                 anno_mark_param$vp_x = unit(0, "npc")
                 anno_mark_param$vp_y = unit(0.5, "npc")
