@@ -25,6 +25,7 @@ setMethod(f = "make_layout",
     definition = function(object) {
 
     if(object@layout$initialized) {
+        qqcat("layout is initialized.\n")
         return(object)
     }
 
@@ -89,19 +90,26 @@ setMethod(f = "make_layout",
     column_title_side = object@column_title_param$side
     column_title_gp = object@column_title_param$gp
     column_title_rot = object@column_title_param$rot
+    if(!is.null(ht_opt$TITLE_PADDING)) {
+        title_padding = ht_opt$TITLE_PADDING
+    } else {
+        title_padding = unit(c(0, 0), "points")
+        title_padding[1] = title_padding[1] + unit(5.5, "points") + 
+            convertHeight(grobDescent(textGrob(label = "jA", gp = column_title_gp)), "inches")
+    }
     if(length(column_title) > 0) {
         if(column_title_side == "top") {
             if(column_title_rot %in% c(0, 180)) {
-                object@layout$layout_size$column_title_top_height = grobHeight(textGrob(column_title, gp = column_title_gp)) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$column_title_top_height = grobHeight(textGrob(column_title, gp = column_title_gp)) + sum(title_padding)
             } else {
-                object@layout$layout_size$column_title_top_height = grobWidth(textGrob(column_title, gp = column_title_gp)) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$column_title_top_height = grobWidth(textGrob(column_title, gp = column_title_gp)) + sum(title_padding)
             }
             object@layout$layout_index = rbind(object@layout$layout_index, column_title_top = heatmap_layout_index("column_title_top"))
         } else {
             if(column_title_rot %in% c(0, 180)) {
-                object@layout$layout_size$column_title_bottom_height = grobHeight(textGrob(column_title, gp = column_title_gp)) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$column_title_bottom_height = grobHeight(textGrob(column_title, gp = column_title_gp)) + sum(title_padding)
             } else {
-                object@layout$layout_size$column_title_bottom_height = grobWidth(textGrob(column_title, gp = column_title_gp)) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$column_title_bottom_height = grobWidth(textGrob(column_title, gp = column_title_gp)) + sum(title_padding)
             }
             object@layout$layout_index = rbind(object@layout$layout_index, column_title_bottom = heatmap_layout_index("column_title_bottom"))
         }
@@ -122,19 +130,26 @@ setMethod(f = "make_layout",
     row_title_side = object@row_title_param$side
     row_title_gp = object@row_title_param$gp
     row_title_rot = object@row_title_param$rot
+    if(!is.null(ht_opt$TITLE_PADDING)) {
+        title_padding = ht_opt$TITLE_PADDING
+    } else {
+        title_padding = unit(c(0, 0), "points")
+        title_padding[1] = title_padding[1] + unit(5.5, "points") + 
+            convertHeight(grobDescent(textGrob(label = "jA", gp = row_title_gp)), "inches")
+    }
     if(length(row_title) > 0) {
         if(row_title_side == "left") {
             if(row_title_rot %in% c(0, 180)) {
-                object@layout$layout_size$row_title_left_width = max_text_width(row_title, gp = row_title_gp) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$row_title_left_width = max_text_width(row_title, gp = row_title_gp) + sum(title_padding)
             } else {
-                object@layout$layout_size$row_title_left_width = max_text_height(row_title, gp = row_title_gp) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$row_title_left_width = max_text_height(row_title, gp = row_title_gp) + sum(title_padding)
             }
             object@layout$layout_index = rbind(object@layout$layout_index, row_title_left = heatmap_layout_index("row_title_left"))
         } else {
             if(row_title_rot %in% c(0, 180)) {
-                object@layout$layout_size$row_title_right_width = max_text_width(row_title, gp = row_title_gp) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$row_title_right_width = max_text_width(row_title, gp = row_title_gp) + sum(title_padding)
             } else {
-                object@layout$layout_size$row_title_right_width = max_text_height(row_title, gp = row_title_gp) + ht_opt$TITLE_PADDING*2
+                object@layout$layout_size$row_title_right_width = max_text_height(row_title, gp = row_title_gp) + sum(title_padding)
             }
             object@layout$layout_index = rbind(object@layout$layout_index, row_title_right = heatmap_layout_index("row_title_right"))
         }
