@@ -63,7 +63,8 @@ HeatmapAnnotation = setClass("HeatmapAnnotation",
 # -annotation_name_gp Graphic parameters for anntation names. Graphic paramters can be vectors.
 # -annotation_name_offset Offset to the annotation names, a `grid::unit` object. The value can be a vector.
 # -annotation_name_side Side of the annotation names.
-# -annotation_name_rot Rotation of the annotation names, it can only take values in ``c(00, 90, 180, 270)``. The value can be a vector.
+# -annotation_name_rot Rotation of the annotation names. The value can be a vector.
+# -annotation_name_align Whether to align the annotation names.
 # -annotation_height Height of each annotation if annotations are column annotations.
 # -annotation_width Width of each annotation if annotations are row annotations.
 # -height Height of the whole column annotations.
@@ -108,6 +109,7 @@ HeatmapAnnotation = function(...,
 	annotation_name_offset = NULL,
 	annotation_name_side = ifelse(which == "column", "right", "bottom"),
 	annotation_name_rot = NULL,
+	annotation_name_align = FALSE,
 	
 	annotation_height = NULL, 
 	annotation_width = NULL, 
@@ -158,7 +160,10 @@ HeatmapAnnotation = function(...,
 		arg_list = scl[-1]
 	} else {
 		scl = as.list(sc[[nsc-1]])
-		if(any(as.character(scl[[1]]) %in% c("HeatmapAnnotation", "rowAnnotation", "columnAnnotation"))) {
+		if(is.function(scl[[1]])) {
+			scl = as.list(sc[[nsc]])
+			arg_list = scl[-1]
+		} else if(any(as.character(scl[[1]]) %in% c("HeatmapAnnotation", "rowAnnotation", "columnAnnotation"))) {
 			arg_list = scl[-1]
 		} else {
 			scl = as.list(sc[[nsc]])
@@ -268,21 +273,34 @@ HeatmapAnnotation = function(...,
     }
     border = recycle_param(border, an, FALSE)
     annotation_name_gp = recycle_gp(annotation_name_gp, n_total_anno)
+    if(length(annotation_label)) {
+    	if(inherits(annotation_label, "gridtext")) {
+    		annotation_label = lapply(seq_along(annotation_label), function(i) annotation_label[i])
+    		names(annotation_label) = an
+    	}
+    }
     annotation_label = recycle_list(annotation_label, an, NULL)
+    # if(length(annotation_label) > 0) {
+    # 	annotation_label = lapply(seq_along(annotation_label), function(i) annotation_label[i])
+    # 	names(annotation_label) == an
+    # }
 
     if(!missing(col)) {
-    	if(is.null(names(col))) {
-    		stop_wrap("`col` should be a named list.")
-    	}
-    	if(any(is.na(names(col)))) {
-    		stop_wrap("`col` should be a named list.")
-    	}
-    	if(any(sapply(col, function(x) if(is.function(x)) FALSE else is.null(names(x))))) {
-    		stop_wrap("elements in `col` should be named vectors.")
-    	}
-    	if(any(sapply(col, function(x) if(is.function(x)) FALSE else any(is.na(names(x)))))) {
-    		stop_wrap("elements in `col` should be named vectors.")
-    	}
+    	if(length(col) >= 1) {
+	    	if(is.null(names(col))) {
+	    		stop_wrap("`col` should be a named list.")
+	    	}
+	    	if(any(is.na(names(col)))) {
+	    		stop_wrap("`col` should be a named list.")
+	    	}
+	    	col = col[intersect(simple_anno_name, names(col))]
+	    	if(any(sapply(col, function(x) if(is.function(x)) FALSE else is.null(names(x))))) {
+	    		stop_wrap("elements in `col` should be named vectors.")
+	    	}
+	    	if(any(sapply(col, function(x) if(is.function(x)) FALSE else any(is.na(names(x)))))) {
+	    		stop_wrap("elements in `col` should be named vectors.")
+	    	}
+	    }
     }
 
     ### check the length of annotations
@@ -314,7 +332,7 @@ HeatmapAnnotation = function(...,
 		i_anno = i_anno + 1
 		arg_list = list(name = ag, which = which,
 				label = annotation_label[[i_anno]],
-				show_name = show_annotation_name[i_anno], 
+				show_name = show_annotation_name[[i_anno]], 
 				name_gp = subset_gp(annotation_name_gp, i_anno), 
 	        	name_offset = annotation_name_offset[[i_anno]], 
 	        	name_side = annotation_name_side[i_anno], 
@@ -435,6 +453,51 @@ HeatmapAnnotation = function(...,
     anno_size = convertWidth(anno_size, "mm")
 
 	names(anno_list) = sapply(anno_list, function(x) x@name)
+
+	if(annotation_name_align) {
+	    # adjust x, y, offset slot in SingleAnnotation objects
+	    l_bottom = sapply(anno_list, function(x) x@name_param$side == "bottom" & x@name_param$show & x@fun@fun_name != "anno_simple")
+	    if(sum(l_bottom) > 1) {
+	    	max_offset = unit(max(sapply(anno_list[l_bottom], function(anno) {
+	    		unit_to_numeric(anno@name_param$offset)
+	    	})), "mm")
+	    	for(i in which(l_bottom)) {
+	    		anno_list[[i]]@name_param$offset = max_offset
+	    		anno_list[[i]]@name_param$y = unit(0, "npc") - max_offset
+	    	}
+	    }
+	    l_top = sapply(anno_list, function(x) x@name_param$side == "top" & x@name_param$show & x@fun@fun_name != "anno_simple")
+	    if(sum(l_top) > 1) {
+	    	max_offset = unit(max(sapply(anno_list[l_top], function(anno) {
+	    		unit_to_numeric(anno@name_param$offset)
+	    	})), "mm")
+	    	for(i in which(l_bottom)) {
+	    		anno_list[[i]]@name_param$offset = max_offset
+	    		anno_list[[i]]@name_param$y = unit(1, "npc") + max_offset
+	    	}
+	    }
+	    l_left = sapply(anno_list, function(x) x@name_param$side == "left" & x@name_param$show & x@fun@fun_name != "anno_simple")
+	    if(sum(l_left) > 1) {
+	    	max_offset = unit(max(sapply(anno_list[l_left], function(anno) {
+	    		unit_to_numeric(anno@name_param$offset)
+	    	})), "mm")
+	    	for(i in which(l_left)) {
+	    		anno_list[[i]]@name_param$offset = max_offset
+	    		anno_list[[i]]@name_param$x = unit(0, "npc") - max_offset
+	    	}
+	    }
+	    l_right = sapply(anno_list, function(x) x@name_param$side == "right" & x@name_param$show & x@fun@fun_name != "anno_simple")
+	    if(sum(l_right) > 1) {
+	    	max_offset = unit(max(sapply(anno_list[l_right], function(anno) {
+	    		unit_to_numeric(anno@name_param$offset)
+	    	})), "mm")
+	    	for(i in which(l_right)) {
+	    		anno_list[[i]]@name_param$offset = max_offset
+	    		anno_list[[i]]@name_param$x = unit(1, "npc") + max_offset
+	    	}
+	    }
+	}
+
     .Object@anno_list = anno_list
     .Object@anno_size = anno_size
     .Object@which = which
@@ -449,6 +512,7 @@ HeatmapAnnotation = function(...,
     	})), "mm")
     }
     .Object@extended = extended
+
     .Object@param = list(
     	simple_anno_size = simple_anno_size, 
     	simple_anno_size_adjust = simple_anno_size_adjust,
@@ -609,8 +673,8 @@ setMethod(f = "draw",
 
     if(test2) {
     	grid.newpage()
-    	if(which == "column") pushViewport(viewport(width = unit(1, "npc") - unit(3, "cm"), height = object@height))
-    	if(which == "row") pushViewport(viewport(height = unit(1, "npc") - unit(3, "cm"), width = object@width))
+    	if(which == "column") pushViewport(viewport(width = unit(1, "npc") - unit(3, "cm") - sum(object@extended[c(2,4)]), height = object@height))
+    	if(which == "row") pushViewport(viewport(height = unit(1, "npc") - unit(3, "cm") - sum(object@extended[c(1,3)]), width = object@width))
     } else {
 		pushViewport(do.call(viewport, vp_param))
 	}
@@ -846,7 +910,7 @@ setMethod(f = "add_heatmap",
 # ha3 = HeatmapAnnotation(sth = cbind(1:10, 10:1))
 # ha = c(ha1, ha2, ha3, gap = unit(c(1, 4), "mm"))
 # ha
-c.HeatmapAnnotation = function(..., gap = unit(0, "mm")) {
+c.HeatmapAnnotation = function(..., gap = unit(1, "points")) {
 	anno_list = list(...)
 	if(length(anno_list) == 1) {
 		return(anno_list[[1]])
@@ -940,6 +1004,9 @@ names.HeatmapAnnotation = function(x) {
 	names(x@anno_list) = value
 	for(i in seq_along(value)) {
 		x@anno_list[[i]]@name =  value[i]
+		if(!is.null(x@anno_list[[i]]@color_mapping)) {
+			x@anno_list[[i]]@color_mapping@name = value[i]
+		}
 	}
 	return(x)
 }
@@ -1353,3 +1420,90 @@ has_zoomed_anno_empty = function(ha) {
 	}
 	return(FALSE)
 }
+
+# == title
+# Attach heatmap annotations to the heatmap
+#
+# == param
+# -object A `Heatmap-class` object.
+# -ha A `HeatmapAnnotation-class` object.
+# -side Which side of the heatmap. Value should be in "top", "bottom", "left", "right".
+# -gap Space between the two heatmap annotations.
+#
+# == example
+# m = matrix(rnorm(100), 10)
+# ht = Heatmap(m)
+# ha = HeatmapAnnotation(foo = 1:10)
+# ht = attach_annotation(ht, ha)
+# ht
+# ha2 = HeatmapAnnotation(bar = letters[1:10])
+# ht = attach_annotation(ht, ha2)
+# ht
+setMethod(f = "attach_annotation",
+    signature = "Heatmap",
+    definition = function(object, ha, side = c("top", "bottom", "left", "right"), 
+    gap = unit(1, "points")) {
+
+    if(missing(side)) {
+    	side = ifelse(ha@which == "column", "top", "left")
+    } else {
+    	side = match.arg(side)[1]
+    }
+    ha_which = ha@which
+    if(ha_which == "column" && side %in% c("left", "right")) {
+    	stop_wrap("Column annotations can only be attached to the top/bottom side of the heatmap.")
+    } else if(ha_which == "row" && side %in% c("top", "bottom")) {
+    	stop_wrap("Row annotations can only be attached to the left/right side of the heatmap.")
+    }
+
+    if(side == "top") {
+    	if(is.null(object@top_annotation)) {
+    		object@top_annotation = ha
+    		h = height(ha) + ht_opt$COLUMN_ANNO_PADDING
+    		h = convertHeight(h, "mm")
+    		object@top_annotation_param = list(height = h)
+    	} else {
+    		object@top_annotation = c(object@top_annotation, ha, gap = gap)
+    		h = height(object@top_annotation) + height(ha) + gap
+    		h = convertHeight(h, "mm")
+    		object@top_annotation_param = list(height = h)
+    	}
+    } else if(side == "bottom") {
+    	if(is.null(object@bottom_annotation)) {
+    		object@bottom_annotation = ha
+    		h = height(ha) + ht_opt$COLUMN_ANNO_PADDING
+    		h = convertHeight(h, "mm")
+    		object@bottom_annotation_param = list(height = h)
+    	} else {
+    		object@bottom_annotation = c(object@bottom_annotation, ha, gap = gap)
+    		h = height(object@bottom_annotation) + height(ha) + gap
+    		h = convertHeight(h, "mm")
+    		object@bottom_annotation_param = list(height = h)
+    	}
+    } else if(side == "left") {
+    	if(is.null(object@left_annotation)) {
+    		object@left_annotation = ha
+    		w = width(ha) + ht_opt$ROW_ANNO_PADDING
+    		w = convertWidth(w, "mm")
+    		object@left_annotation_param = list(width = w)
+    	} else {
+    		object@left_annotation = c(object@left_annotation, ha, gap = gap)
+    		w = width(object@left_annotation) + width(ha) + gap
+    		w = convertWidth(w, "mm")
+    		object@left_annotation_param = list(width = w)
+    	}
+    } else if(side == "right") {
+    	if(is.null(object@right_annotation)) {
+    		object@right_annotation = ha
+    		w = width(ha) + ht_opt$ROW_ANNO_PADDING
+    		w = convertWidth(w, "mm")
+    		object@right_annotation_param = list(width = w)
+    	} else {
+    		object@right_annotation = c(object@right_annotation, ha, gap = gap)
+    		w = width(object@right_annotation) + width(ha) + gap
+    		w = convertWidth(w, "mm")
+    		object@right_annotation_param = list(height = w)
+    	}
+    }
+    object
+})
