@@ -103,7 +103,7 @@ setMethod(f = "add_heatmap",
     
     direction = match.arg(direction)[1]
     if(object@direction != direction) {
-        stop_wrap("The heatmap list should only be all horizontal or vertical.")
+        stop_wrap("The heatmap list should only be all horizontal or vertical. Maybe you can move some of the row annotations to the 'left_annotation'/'right_annotation' of the heatmap if you want to make the heatmap list vertical. Or similarly, to move independent column annotations to 'top_annotation'/'bottom_annotation' if you want the heatmap list horizontal.")
     }
 
     # check settings of this new heatmap
@@ -134,30 +134,61 @@ setMethod(f = "add_heatmap",
         warning_wrap(paste0("Heatmap/annotation names are duplicated: ", paste(ht_name[which_duplicated], collapse = ", ")))
     }
 
-    l = which(sapply(object@ht_list, inherits, "Heatmap"))
     if(direction == "horizontal") {
-        nr = sapply(object@ht_list[l], function(ht) nrow(ht@matrix))
+        nr = sapply(object@ht_list, function(ht) {
+            if(inherits(ht, "Heatmap")) {
+                nrow(ht@matrix)
+            } else if(inherits(ht, "HeatmapAnnotation")) {
+                nobs(ht)
+            } else {
+                NA
+            }
+        })
+        nr = nr[!is.na(nr)]
 
         if(length(unique(nr)) > 1) {
-            cat("`nrow` of all heatmaps:\n")
-            print(nr)
-            stop_wrap("`nrow` of all heatmaps should be the same for horizontal heatmap list.")
-            for(i in l) {
-                cat(object@ht_list[[i]]@name, ":", nrow(object@ht_list[[i]]@matrix), "\n")
+            cat("`nrow` of all heatmaps and `nobs` of all annotations should be the same\nfor horizontal heatmap list.\n")
+            for(i in seq_along(object@ht_list)) {
+                if(inherits(object@ht_list[[i]], "Heatmap")) {
+                    cat("  heatmap '", object@ht_list[[i]]@name, "': ", nrow(object@ht_list[[i]]@matrix), "\n", sep = "")
+                } else if(inherits(object@ht_list[[i]], "HeatmapAnnotation")) {
+                    ha = object@ht_list[[i]]
+                    for(j in seq_along(ha@anno_list)) {
+                        if(!is.na(nobs(ha@anno_list[[j]]))) {
+                            cat("  annotation '", ha@anno_list[[j]]@name, "': ", nobs(ha@anno_list[[j]]), "\n", sep = "")
+                        }
+                    }
+                }
             }
-            cat("\n")
+            stop("You have an error when adding heatmaps and annotations.")
         }
     } else {
-        nc = sapply(object@ht_list[l], function(ht) ncol(ht@matrix))
+        nc = sapply(object@ht_list, function(ht) {
+            if(inherits(ht, "Heatmap")) {
+                ncol(ht@matrix)
+            } else if(inherits(ht, "HeatmapAnnotation")) {
+                nobs(ht)
+            } else {
+                NA
+            }
+        })
+        nc = nc[!is.na(nc)]
 
         if(length(unique(nc)) > 1) {
-            cat("`ncol` of all heatmaps:\n")
-            print(nc)
-            stop_wrap("`ncol` of all heatmaps should be the same for vertical heatmap list.")
-            for(i in l) {
-                cat(object@ht_list[[i]]@name, ":", ncol(object@ht_list[[i]]@matrix), "\n")
+            cat("`ncol` of all heatmaps and `nobs` of all annotations should be the same\nfor vertical heatmap list.\n")
+            for(i in seq_along(object@ht_list)) {
+                if(inherits(object@ht_list[[i]], "Heatmap")) {
+                    cat("  heatmap '", object@ht_list[[i]]@name, "': ", ncol(object@ht_list[[i]]@matrix), "\n", sep = "")
+                } else if(inherits(object@ht_list[[i]], "HeatmapAnnotation")) {
+                    ha = object@ht_list[[i]]
+                    for(j in seq_along(ha@anno_list)) {
+                        if(!is.na(nobs(ha@anno_list[[j]]))) {
+                            cat("  annotation '", ha@anno_list[[j]]@name, "': ", nobs(ha@anno_list[[j]]), "\n", sep = "")
+                        }
+                    }
+                }
             }
-            cat("\n")
+            stop("You have an error when adding heatmaps and annotations.")
         }
     }
 
@@ -184,6 +215,11 @@ setMethod(f = "add_heatmap",
 # -annotation_legend_side side of the annotation legends
 # -show_annotation_legend whether show annotation legends
 # -annotation_legend_list user-defined legends which are put after the annotation legends
+# -align_heatmap_legend How to align the legends to heatmap. Possible values are "heatmap_center", "heatmap_top" and "global_center". If the value is ``NULL``,
+#           it automatically picks the proper value from the three options.
+# -align_annotation_legend  How to align the legends to heatmap. Possible values are "heatmap_center", "heatmap_top" and "global_center".
+# -legend_grouping How the legends are grouped. Values should be "adjusted" or "original". If it is set as
+#             "original", all annotation legends are grouped together.
 # -gap gap between heatmaps/annotations
 # -ht_gap same as ``gap``.
 # -main_heatmap index of main heatmap. The value can be a numeric index or the heatmap name
@@ -272,10 +308,10 @@ setMethod(f = "draw",
 
     row_title = character(0), 
     row_title_side = c("left", "right"), 
-    row_title_gp = gpar(fontsize = 14),
+    row_title_gp = gpar(fontsize = 13),
     column_title = character(0), 
     column_title_side = c("top", "bottom"), 
-    column_title_gp = gpar(fontsize = 14), 
+    column_title_gp = gpar(fontsize = 13), 
 
     heatmap_legend_side = c("right", "left", "bottom", "top"), 
     merge_legends = FALSE,
@@ -284,6 +320,9 @@ setMethod(f = "draw",
     annotation_legend_side = c("right", "left", "bottom", "top"), 
     show_annotation_legend = TRUE, 
     annotation_legend_list = list(),
+    align_heatmap_legend = NULL,
+    align_annotation_legend = NULL,
+    legend_grouping = c("adjusted", "original"),
 
     gap = unit(2, "mm"), 
     ht_gap = gap, 
@@ -359,6 +398,13 @@ setMethod(f = "draw",
 
     verbose = ht_opt$verbose
 
+    if(missing(cluster_rows) && !missing(row_order)) {
+        cluster_rows = FALSE
+    }
+    if(missing(cluster_columns) && !missing(column_order)) {
+        cluster_columns = FALSE
+    }
+
     ovl = list()
     for(opt_nm in c("heatmap_row_names_gp",
                     "heatmap_column_names_gp",
@@ -440,17 +486,15 @@ setMethod(f = "draw",
         grid.newpage()
     }
 
-    # if(!missing(heatmap_legend_offset) && missing(annotation_legend_offset)) {
-    #     annotation_legend_offset = heatmap_legend_offset
-    # } else if(missing(heatmap_legend_offset) && !missing(annotation_legend_offset)) {
-    #     heatmap_legend_offset = annotation_legend_offset
-    # }
-
-    if(is.null(object@heatmap_legend_param$offset))
-        object@heatmap_legend_param$offset = unit(0, "mm") #heatmap_legend_offset
-    if(is.null(object@annotation_legend_param$offset))
-        object@annotation_legend_param$offset = unit(0, "mm") #annotation_legend_offset
-
+    heatmap_legend_side = match.arg(heatmap_legend_side)
+    annotation_legend_side = match.arg(annotation_legend_side)
+    if(heatmap_legend_side == annotation_legend_side) {
+        if(!is.null(align_heatmap_legend) && is.null(align_annotation_legend)) {
+            align_annotation_legend = align_heatmap_legend
+        } else if(is.null(align_heatmap_legend) && !is.null(align_annotation_legend)) {
+            align_heatmap_legend = align_heatmap_legend
+        }
+    }
     object@ht_list_param$adjust_annotation_extension = adjust_annotation_extension
 
     object = make_layout(
@@ -469,6 +513,9 @@ setMethod(f = "draw",
         annotation_legend_side = annotation_legend_side, 
         show_annotation_legend = show_annotation_legend, 
         annotation_legend_list = annotation_legend_list,
+        align_heatmap_legend = align_heatmap_legend,
+        align_annotation_legend = align_annotation_legend,
+        legend_grouping = legend_grouping,
 
         ht_gap = ht_gap, 
 
@@ -520,54 +567,302 @@ setMethod(f = "draw",
         raster_resize = raster_resize
     )
 
-    ht_list_width = sum(component_width(object)) + padding[2] + padding[4]
-    ht_list_height = sum(component_height(object)) + padding[1] + padding[3]
+    code = expression({
 
-    if(is_abs_unit(ht_list_width)) {
-        ht_list_width = unit(convertWidth(ht_list_width, "mm", valueOnly = TRUE), "mm")
-        # qqcat("Since all heatmaps/annotations have absolute units, the total width of the plot is @{ht_list_width}\n")
-        w = ht_list_width
-    } else {
-        w = unit(1, "npc")
-    }
-    if(is_abs_unit(ht_list_height)) {
-        ht_list_height = unit(convertHeight(ht_list_height, "mm", valueOnly = TRUE), "mm")
-        # qqcat("Since all heatmaps/annotations have absolute units, the total height of the plot is @{ht_list_height}\n")
-        h = ht_list_height
-    } else {
-        h = unit(1, "npc")
-    }
+        current_vp = current.viewport()$name
+        if(current_vp == "ROOT") {
+            page_size = unit(par("din"), "in")
+        } else {
+            grid::upViewport()
+            page_size = unit.c(convertWidth(unit(1, "npc"), "mm"),
+                                             convertHeight(unit(1, "npc"), "mm"))
+            grid::downViewport(current_vp)
+        }
 
-    layout = grid.layout(nrow = length(HEATMAP_LIST_LAYOUT_COLUMN_COMPONENT), 
-        ncol = length(HEATMAP_LIST_LAYOUT_ROW_COMPONENT), 
-        widths = component_width(object), 
-        heights = component_height(object))
+        # adjust legend, the following have 99% duplication as in HeatmapList-layout.R
+        # however, it runs fast, so I just make it ugly
 
-    pushViewport(viewport(name = "global", width = w, height = h))
-    pushViewport(viewport(layout = layout, name = "global_layout", x = padding[2], y = padding[1], width = unit(1, "npc") - padding[2] - padding[4],
-        height = unit(1, "npc") - padding[1] - padding[3], just = c("left", "bottom")))
-    ht_layout_index = object@layout$layout_index
-    ht_graphic_fun_list = object@layout$graphic_fun_list
-
-    for(j in seq_len(nrow(ht_layout_index))) {
-        pushViewport(viewport(name = paste0("global-", rownames(ht_layout_index)[j]), layout.pos.row = ht_layout_index[j, 1], layout.pos.col = ht_layout_index[j, 2]))
-        ht_graphic_fun_list[[j]](object)
-        upViewport()
-    }
-
-    upViewport()
-    upViewport()
-
-    object@ht_list_param$width = w
-    object@ht_list_param$height = h
-
-    for(i in seq_along(object@ht_list)) {
-        if(inherits(object@ht_list[[i]], "Heatmap")) {
+        #################################################
+        ## heatmap legend to top, bottom, left and right
+        # default values
+        legend_grouping = object@ht_list_param$legend_grouping
+        ColorMappingList = list()
+        for(i in seq_along(object@ht_list)) {
             ht = object@ht_list[[i]]
-            if(!is.null(ht@heatmap_param$post_fun)) {
-                ht@heatmap_param$post_fun(ht)
+            if(direction == "horizontal") {
+                if(inherits(object@ht_list[[i]], "Heatmap")) {
+                    if(!is.null(ht@left_annotation)) {
+                        if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@left_annotation))
+                        }
+                    }
+                    if(!is.null(ht@top_annotation)) {
+                        if(object@ht_list_param$merge_legends) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@top_annotation))
+                        }
+                    }
+                    if(object@ht_list[[i]]@heatmap_param$show_heatmap_legend) {
+                        ColorMappingList = c.list(ColorMappingList, object@ht_list[[i]]@matrix_color_mapping)
+                    }
+                    if(!is.null(ht@bottom_annotation)) {
+                        if(object@ht_list_param$merge_legends) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@bottom_annotation))
+                        }
+                    }
+                    if(!is.null(ht@right_annotation)) {
+                        if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@right_annotation))
+                        }
+                    }
+                } else if(inherits(object@ht_list[[i]], "HeatmapAnnotation")) {
+                    if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                        ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(object@ht_list[[i]]))
+                    }
+                }
+            } else {
+                if(inherits(object@ht_list[[i]], "Heatmap")) {
+                    if(!is.null(ht@left_annotation)) {
+                        if(object@ht_list_param$merge_legends) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@left_annotation))
+                        }
+                    }
+                    if(!is.null(ht@top_annotation)) {
+                        if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@top_annotation))
+                        }
+                    }
+                    if(object@ht_list[[i]]@heatmap_param$show_heatmap_legend) {
+                        ColorMappingList = c.list(ColorMappingList, object@ht_list[[i]]@matrix_color_mapping)
+                    }
+                    if(!is.null(ht@bottom_annotation)) {
+                        if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@bottom_annotation))
+                        }
+                    }
+                    if(!is.null(ht@right_annotation)) {
+                        if(object@ht_list_param$merge_legends) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@right_annotation))
+                        }
+                    }
+                } else if(inherits(object@ht_list[[i]], "HeatmapAnnotation")) {
+                    if(object@ht_list_param$merge_legends || legend_grouping == "adjusted") {
+                        ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(object@ht_list[[i]]))
+                    }
+                }
             }
         }
+        if(length(heatmap_legend_list) != 0) {
+            if(inherits(heatmap_legend_list, c("Legends", "grob"))) {
+                heatmap_legend_list = list(heatmap_legend_list)
+            }
+        }
+        if(length(annotation_legend_list) != 0) {
+            if(inherits(annotation_legend_list, c("Legends", "grob"))) {
+                annotation_legend_list = list(annotation_legend_list)
+            }
+        }
+        if(merge_legends) {
+            heatmap_legend_list = c(heatmap_legend_list, annotation_legend_list)
+        }
+        if(length(ColorMappingList) == 0 && length(heatmap_legend_list) == 0) {
+            show_heatmap_legend = FALSE
+        }
+
+        object@heatmap_legend_param$show = show_heatmap_legend
+        heatmap_legend_side = object@heatmap_legend_param$side
+        if(show_heatmap_legend) {
+            if(heatmap_legend_side == "top") {
+                object@heatmap_legend_param$padding = unit.c(ht_opt$HEATMAP_LEGEND_PADDING, unit(c(0, 0, 0), "mm")) #unit(c(2, 0, 0, 0), "mm")
+                size = heatmap_legend_size(object, legend_list = heatmap_legend_list, max_width = calc_legends_max_width(object, page_size))
+                object@heatmap_legend_param$size = size
+                object@layout$layout_heatmap_legend_top_height = size[2]
+                layout_index = heatmap_list_layout_index("heatmap_legend_top")
+            } else if(heatmap_legend_side == "bottom") {
+                object@heatmap_legend_param$padding = unit.c(unit(c(0, 0), "mm"), ht_opt$HEATMAP_LEGEND_PADDING, unit(0, "mm")) # unit(c(0, 0, 2, 0), "mm")
+                size = heatmap_legend_size(object, legend_list = heatmap_legend_list, max_width = calc_legends_max_width(object, page_size))
+                object@heatmap_legend_param$size = size
+                object@layout$layout_heatmap_legend_bottom_height = size[2]
+                layout_index = heatmap_list_layout_index("heatmap_legend_bottom")
+            } else if(heatmap_legend_side == "left") {
+                object@heatmap_legend_param$padding = unit.c(unit(c(0, 0, 0), "mm"), ht_opt$HEATMAP_LEGEND_PADDING) # unit(c(0, 0, 0, 2), "mm")
+                size = heatmap_legend_size(object, legend_list = heatmap_legend_list, max_height = calc_legends_max_height(object, page_size))
+                object@heatmap_legend_param$size = size
+                object@layout$layout_heatmap_legend_left_width = size[1]
+                layout_index = heatmap_list_layout_index("heatmap_legend_left")
+            } else if(heatmap_legend_side == "right") {
+                object@heatmap_legend_param$padding = unit.c(unit(0, "mm"), ht_opt$HEATMAP_LEGEND_PADDING, unit(c(0, 0), "mm")) # unit(c(0, 2, 0, 0), "mm")
+                size = heatmap_legend_size(object, legend_list = heatmap_legend_list, max_height = calc_legends_max_height(object, page_size))
+                object@heatmap_legend_param$size = size
+                object@layout$layout_heatmap_legend_right_width = size[1]
+                layout_index = heatmap_list_layout_index("heatmap_legend_right")
+            }
+            if(heatmap_legend_side %in% c("top", "bottom")) {
+                ind = which(object@layout$layout_index[, 1] == layout_index[1] & object@layout$layout_index[, 2] == layout_index[2])
+                object@layout$graphic_fun_list[[ind]] = function(object) draw_heatmap_legend(object, legend_list = heatmap_legend_list, max_width = calc_legends_max_width(object, page_size))
+            } else {
+                ind = which(object@layout$layout_index[, 1] == layout_index[1] & object@layout$layout_index[, 2] == layout_index[2])
+                object@layout$graphic_fun_list[[ind]] = function(object) draw_heatmap_legend(object, legend_list = heatmap_legend_list, max_height = calc_legends_max_height(object, page_size))
+            }
+        } else {
+            object@heatmap_legend_param$size = unit(c(0, 0), "mm")
+        }
+
+        #################################################
+        ## annotation legend to top, bottom, left and right
+        # default values
+        ColorMappingList = list()
+        if(!merge_legends) {
+            for(i in seq_along(object@ht_list)) {
+                ht = object@ht_list[[i]]
+                if(direction == "horizontal") {
+                    if(inherits(ht, "Heatmap")) {
+                        if(!is.null(ht@left_annotation)) {
+                            if(legend_grouping == "original") {
+                                ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@left_annotation))
+                            }
+                        }
+                        if(!is.null(ht@top_annotation)) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@top_annotation))
+                        }
+                        if(!is.null(ht@bottom_annotation)) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@bottom_annotation))
+                        }
+                        if(!is.null(ht@right_annotation)) {
+                            if(legend_grouping == "original") {
+                                ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@right_annotation))
+                            }
+                        }
+                    } else if(inherits(ht, "HeatmapAnnotation")) {
+                        if(legend_grouping == "original") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht))
+                        }
+                    }
+                } else {
+                    if(inherits(ht, "Heatmap")) {
+                        if(!is.null(ht@left_annotation)) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@left_annotation))
+                        }
+                        if(!is.null(ht@top_annotation)) {
+                            if(legend_grouping == "original") {
+                                ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@top_annotation))
+                            }
+                        }
+                        if(!is.null(ht@bottom_annotation)) {
+                            if(legend_grouping == "original") {
+                                ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@bottom_annotation))
+                            }
+                        }
+                        if(!is.null(ht@top_annotation)) {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht@top_annotation))
+                        }
+                    } else if(inherits(ht, "HeatmapAnnotation")) {
+                        if(legend_grouping == "original") {
+                            ColorMappingList = c.list(ColorMappingList, list = get_color_mapping_list(ht))
+                        }
+                    }
+                }
+            }
+        } else {
+            annotation_legend_list = list()
+        }
+        if(length(ColorMappingList) == 0 && length(annotation_legend_list) == 0) {
+            show_annotation_legend = FALSE
+        }
+        object@annotation_legend_param$show = show_annotation_legend
+        annotation_legend_side = object@annotation_legend_param$side
+        if(show_annotation_legend) {
+            if(annotation_legend_side == "top") {
+                object@annotation_legend_param$padding = unit.c(ht_opt$ANNOTATION_LEGEND_PADDING, unit(c(0, 0, 0), "mm")) # unit(c(2, 0, 0, 0), "mm")
+                size = annotation_legend_size(object, legend_list = annotation_legend_list, max_width = calc_legends_max_width(object, page_size))
+                object@annotation_legend_param$size = size
+                object@layout$layout_annotation_legend_top_height = size[2]
+                layout_index = heatmap_list_layout_index("annotation_legend_top")
+            } else if(annotation_legend_side == "bottom") {
+                object@annotation_legend_param$padding = unit.c(unit(c(0, 0), "mm"), ht_opt$ANNOTATION_LEGEND_PADDING, unit(0, "mm")) # unit(c(0, 0, 2, 0), "mm")
+                size = annotation_legend_size(object, legend_list = annotation_legend_list, max_width = calc_legends_max_width(object, page_size))
+                object@annotation_legend_param$size = size
+                object@layout$layout_annotation_legend_bottom_height = size[2]
+                layout_index = heatmap_list_layout_index("annotation_legend_bottom")
+            } else if(annotation_legend_side == "left") {
+                object@annotation_legend_param$padding = unit.c(unit(c(0, 0, 0), "mm"), ht_opt$ANNOTATION_LEGEND_PADDING) # unit(c(0, 0, 0, 2), "mm")
+                size = annotation_legend_size(object, legend_list = annotation_legend_list, max_height = calc_legends_max_height(object, page_size))
+                object@annotation_legend_param$size = size
+                object@layout$layout_annotation_legend_left_width = size[1]
+                layout_index = heatmap_list_layout_index("annotation_legend_left")
+            } else if(annotation_legend_side == "right") {
+                object@annotation_legend_param$padding = unit.c(unit(0, "mm"), ht_opt$ANNOTATION_LEGEND_PADDING, unit(c(0, 0), "mm")) # unit(c(0, 2, 0, 0), "mm")
+                size = annotation_legend_size(object, legend_list = annotation_legend_list, max_height = calc_legends_max_height(object, page_size))
+                object@annotation_legend_param$size = size
+                object@layout$layout_annotation_legend_right_width = size[1]
+                layout_index = heatmap_list_layout_index("annotation_legend_right")
+            }
+            if(annotation_legend_side %in% c("top", "bottom")) {
+                ind = which(object@layout$layout_index[, 1] == layout_index[1] & object@layout$layout_index[, 2] == layout_index[2])
+                object@layout$graphic_fun_list[[ind]] = function(object) draw_annotation_legend(object, legend_list = annotation_legend_list, max_width = calc_legends_max_width(object, page_size))
+            } else {
+                ind = which(object@layout$layout_index[, 1] == layout_index[1] & object@layout$layout_index[, 2] == layout_index[2])
+                object@layout$graphic_fun_list[[ind]] = function(object) draw_annotation_legend(object, legend_list = annotation_legend_list, max_height = calc_legends_max_height(object, page_size))
+            }
+        } else {
+            object@annotation_legend_param$size = unit(c(0, 0), "null")
+        }
+
+        ht_list_width = sum(component_width(object)) + padding[2] + padding[4]
+        ht_list_height = sum(component_height(object)) + padding[1] + padding[3]
+
+        if(is_abs_unit(ht_list_width)) {
+            ht_list_width = unit(convertWidth(ht_list_width, "mm", valueOnly = TRUE), "mm")
+            # qqcat("Since all heatmaps/annotations have absolute units, the total width of the plot is @{ht_list_width}\n")
+            w = ht_list_width
+        } else {
+            w = unit(1, "npc")
+        }
+        if(is_abs_unit(ht_list_height)) {
+            ht_list_height = unit(convertHeight(ht_list_height, "mm", valueOnly = TRUE), "mm")
+            # qqcat("Since all heatmaps/annotations have absolute units, the total height of the plot is @{ht_list_height}\n")
+            h = ht_list_height
+        } else {
+            h = unit(1, "npc")
+        }
+
+        layout = grid.layout(nrow = length(HEATMAP_LIST_LAYOUT_COLUMN_COMPONENT), 
+            ncol = length(HEATMAP_LIST_LAYOUT_ROW_COMPONENT), 
+            widths = component_width(object), 
+            heights = component_height(object))
+
+        pushViewport(viewport(name = "global", width = w, height = h, gp = gpar(lineheight = 0.9)))
+        grid.rect(gp = gpar(fill = "white", col = "white"))
+        pushViewport(viewport(layout = layout, name = "global_layout", x = padding[2], y = padding[1], width = unit(1, "npc") - padding[2] - padding[4],
+            height = unit(1, "npc") - padding[1] - padding[3], just = c("left", "bottom")))
+        ht_layout_index = object@layout$layout_index
+        ht_graphic_fun_list = object@layout$graphic_fun_list
+
+        for(j in seq_len(nrow(ht_layout_index))) {
+            pushViewport(viewport(name = paste0("global-", rownames(ht_layout_index)[j]), layout.pos.row = ht_layout_index[j, 1], layout.pos.col = ht_layout_index[j, 2]))
+            ht_graphic_fun_list[[j]](object)
+            upViewport()
+        }
+
+        upViewport()
+        upViewport()
+
+        object@ht_list_param$width = w
+        object@ht_list_param$height = h
+
+        for(i in seq_along(object@ht_list)) {
+            if(inherits(object@ht_list[[i]], "Heatmap")) {
+                ht = object@ht_list[[i]]
+                if(!is.null(ht@heatmap_param$post_fun)) {
+                    ht@heatmap_param$post_fun(ht)
+                }
+            }
+        }
+    })
+    
+    if(is_RStudio_current_dev() || (!dev.interactive())) {
+        eval(code)
+    } else {
+        grDevices::recordGraphics(eval(code), list(), as.environment(-1))
     }
 
     return(invisible(object))
@@ -804,3 +1099,35 @@ summary.HeatmapList = function(object, ...) {
     }
 }
 
+# == title
+# Calculate the width and height of the heatmaps
+#
+# == param
+# -ht A `Heatmap-class` or `HeatmapList-class` object.
+#
+# == value
+# A list of two elements: width and height.
+#
+ht_size = function(ht) {
+    dev.null()
+    ht = draw(ht)
+    w = width(ht)
+    h = height(ht)
+    dev.off2()
+
+    list(width = w, height = h)
+}
+
+
+which_first_ht = function(ht_list) {
+    which(sapply(ht_list@ht_list, inherits, "Heatmap"))[1]
+}
+
+
+which_last_ht = function(ht_list) {
+    max(which(sapply(ht_list@ht_list, inherits, "Heatmap")))
+}
+
+which_main_ht = function(ht_list) {
+    ht_list@ht_list_param$main_heatmap
+}
