@@ -775,6 +775,14 @@ setMethod(f = "draw_title",
         "row" = object@row_title,
         "column" = object@column_title)
 
+    if(!is.null(ht_opt$TITLE_PADDING)) {
+        title_padding = ht_opt$TITLE_PADDING
+    } else {
+        title_padding = unit(c(0, 0), "points")
+        title_padding[1] = title_padding[1] + unit(5.5, "points") + 
+            convertHeight(grobDescent(textGrob(label = "jA", gp = gp)), "inches")
+    }
+
     if(which == "row") {
         rot = switch(side,
             "left" = 90,
@@ -784,14 +792,22 @@ setMethod(f = "draw_title",
         if("fill" %in% names(gp)) {
             grid.rect(gp = gpar(fill = gp$fill))
         }
-        grid.text(title, rot = rot, gp = gp)
+        if(side == "left") {
+            grid.text(title, x = unit(1, "npc") - title_padding[1], rot = rot, just = "bottom", gp = gp)
+        } else {
+            grid.text(title, x = title_padding[1], rot = rot, just = "bottom", gp = gp)
+        }
         upViewport()
     } else {
         pushViewport(viewport(name = "global_column_title"))
         if("fill" %in% names(gp)) {
             grid.rect(gp = gpar(fill = gp$fill))
         }
-        grid.text(title, gp = gp)
+        if(side == "top") {
+            grid.text(title, y = title_padding[1], just = "bottom", gp = gp)
+        } else {
+            grid.text(title, y = unit(1, "npc") - title_padding[1], just = "top", gp = gp)
+        }
         upViewport()
     }
 })
