@@ -24,6 +24,8 @@ make_comb_mat_from_matrix = function(x, mode, top_n_sets = Inf, min_set_size = -
 				FALSE
 			}
 		})
+	} else {
+		stop_wrap("The input should be a matrix or in a format that can be converted to a matrix.")
 	}
 	if(sum(lc) < 1) {
 		stop_wrap("Can not find columns which are logical or only contain 0 or 1.")
@@ -35,6 +37,10 @@ make_comb_mat_from_matrix = function(x, mode, top_n_sets = Inf, min_set_size = -
 	}
 
 	x = as.matrix(x) + 0
+	if(any(is.na(x))) {
+		warning_wrap("The matrix contains NA values. Convert to 0/FALSE.")
+		x[is.na(x)] = 0
+	}
 
 	if(is.null(rownames(x))) {
 		x_has_rownames = FALSE
@@ -323,7 +329,7 @@ list_to_matrix = function(lt, universal_set = NULL) {
 	rownames(mat) = sort(universal_set)
 	colnames(mat) = names(lt)
 	for(i in seq_along(lt)) {
-		mat[unique(lt[[i]]), i] = 1
+		mat[as.character(unique(lt[[i]])), i] = 1
 	}
 	return(mat)
 }
@@ -444,6 +450,11 @@ make_comb_mat = function(..., mode = c("distinct", "intersect", "union"),
 	if(length(lt) == 1) {
 		lt = lt[[1]]
 		if(length(dim(lt)) == 2) { # a matrix
+
+			if(ncol(lt) > 31) {
+				stop_wrap("Only support number of sets <= 31.")
+			}
+
 			m = make_comb_mat_from_matrix(lt, mode = mode, top_n_sets = top_n_sets, 
 				min_set_size = min_set_size, universal_set = universal_set, complement_size = complement_size,
 				set_on_rows = set_on_rows)
@@ -461,7 +472,11 @@ make_comb_mat = function(..., mode = c("distinct", "intersect", "union"),
 		}
 	}
 
-	# if lt is a list of atomic sets, convert to the matrix because it is more efficiency
+	if(length(lt) > 31) {
+		stop_wrap("Only support number of sets <= 31.")
+	}
+
+	# if lt is a list of atomic sets, convert to the matrix because it is more efficient
 	if(is.atomic(lt[[1]])) {
 		m = make_comb_mat_from_matrix(list_to_matrix(lt), mode = mode, top_n_sets = top_n_sets, 
 			min_set_size = min_set_size, universal_set = universal_set, complement_size = complement_size,
@@ -1489,6 +1504,10 @@ order.comb_mat = function(m, decreasing = TRUE, on = "comb_set") {
 # The default top annotation is actually barplot implemented by `anno_barplot`. For
 # how to set the top annotation or bottom annotation in `UpSet`, please refer to `UpSet`.
 #
+# If you want to use `decorate_annotation` function, the annotation name for the "sets"
+# is ``set_size`` and the annotation name for the "intersection sets" are ``intersection_size``
+# and if under the union mode, it is ``union_size``.
+#
 upset_top_annotation = function(m, 
 	gp = gpar(fill = "black"), 
 	height = unit(ifelse(set_on_rows, 3, 2), "cm"),
@@ -1502,29 +1521,32 @@ upset_top_annotation = function(m,
 	set_on_rows = attr(m, "param")$set_on_rows
 	
 	if(set_on_rows) {
-		ha = HeatmapAnnotation("Intersection\nsize" = anno_barplot(comb_size(m), 
+		ha = HeatmapAnnotation("intersection_size" = anno_barplot(comb_size(m), 
 					border = FALSE, gp = gp, height = height, ...), 
 				show_annotation_name = show_annotation_name,
 				annotation_name_gp = annotation_name_gp,
 				annotation_name_offset = annotation_name_offset,
 				annotation_name_side = annotation_name_side,
-				annotation_name_rot = annotation_name_rot)
+				annotation_name_rot = annotation_name_rot,
+				annotation_label = "Intersection\nsize")
 	} else {
-		ha = HeatmapAnnotation("Set\nsize" = anno_barplot(set_size(m), border = FALSE, 
+		ha = HeatmapAnnotation("set_size" = anno_barplot(set_size(m), border = FALSE, 
 					gp = gp, height = height, ...),
 				show_annotation_name = show_annotation_name,
 				annotation_name_gp = annotation_name_gp,
 				annotation_name_offset = annotation_name_offset,
 				annotation_name_side = annotation_name_side,
-				annotation_name_rot = annotation_name_rot)
+				annotation_name_rot = annotation_name_rot,
+				annotation_label = "Set\nsize")
 	}
 
 	mode = attr(m, "param")$mode
 	if(set_on_rows) {
 		if(mode %in% c("distinct", "intersect")) {
-			names(ha) = "Intersection\nsize"
+			# names(ha) = "intersection_size"
 		} else {
-			names(ha) = "Union\nsize"
+			names(ha) = "union_size"
+			ha@anno_list[[1]]@label = "Union\nsize"
 		}
 	}
 	return(ha)
@@ -1548,6 +1570,10 @@ upset_top_annotation = function(m,
 # The default right annotation is actually barplot implemented by `anno_barplot`. For
 # how to set the right annotation or left annotation in `UpSet`, please refer to `UpSet`.
 #
+# If you want to use `decorate_annotation` function, the annotation name for the "sets"
+# is ``set_size`` and the annotation name for the "intersection sets" are ``intersection_size``
+# and if under the union mode, it is ``union_size``.
+#
 upset_right_annotation = function(m,
 	gp = gpar(fill = "black"),  
 	width = unit(ifelse(set_on_rows, 2, 3), "cm"),
@@ -1561,29 +1587,32 @@ upset_right_annotation = function(m,
 	set_on_rows = attr(m, "param")$set_on_rows
 
 	if(set_on_rows) {
-		ha = rowAnnotation("Set size" = anno_barplot(set_size(m), border = FALSE, 
+		ha = rowAnnotation("set_size" = anno_barplot(set_size(m), border = FALSE, 
 					gp = gp, width = width, ...),
 				show_annotation_name = show_annotation_name,
 				annotation_name_gp = annotation_name_gp,
 				annotation_name_offset = annotation_name_offset,
 				annotation_name_side = annotation_name_side,
-				annotation_name_rot = annotation_name_rot)
+				annotation_name_rot = annotation_name_rot,
+				annotation_label = "Set size")
 	} else {
-		ha = rowAnnotation("Intersection\nsize" = anno_barplot(comb_size(m), 
+		ha = rowAnnotation("intersection_size" = anno_barplot(comb_size(m), 
 					border = FALSE, gp = gp, width = width, ...),
 				show_annotation_name = show_annotation_name,
 				annotation_name_gp = annotation_name_gp,
 				annotation_name_offset = annotation_name_offset,
 				annotation_name_side = annotation_name_side,
-				annotation_name_rot = annotation_name_rot)
+				annotation_name_rot = annotation_name_rot,
+				annotation_label = "Intersection\nsize")
 	}
 
 	mode = attr(m, "param")$mode
 	if(!set_on_rows) {
 		if(mode %in% c("distinct", "intersect")) {
-			names(ha) = "Intersection\nsize"
+			# names(ha) = "intersection_size"
 		} else {
-			names(ha) = "Union size"
+			names(ha) = "union_size"
+			ha@anno_list[[1]]@label = "Union size"
 		}
 	}
 	return(ha)
@@ -1593,31 +1622,35 @@ upset_right_annotation = function(m,
 # Normalize a list of combination matrice
 #
 # == param
-# -... If it is a single argument, the value should be a list of combination matrices.
+# -... Combination matrices.
+# -full_comb_sets Whether the combination matrices contain the full sets of combination sets?
+# -complement_set Whether the combination matrices also contain the complement set?
 #
 # == details
 # It normalizes a list of combination matrice to make them have same number and order of sets and combination sets.
 #
 # The sets (by `set_name`) from all combination matrice should be the same.
 #
-normalize_comb_mat = function(...) {
+normalize_comb_mat = function(..., full_comb_sets = FALSE, complement_set = FALSE) {
+
 	lt = list(...)
+	input_is_comb_mat = FALSE
 	if(length(lt) == 1) {
-		if(!is.list(lt)) {
-			stop_wrap("If you only specify one argument, it must be a list of comb_mat objects.")
+		if(inherits(lt[[1]], "comb_mat")) {
+			input_is_comb_mat = TRUE
+		}
+		if(is.list(lt[[1]])) {
+			lt = lt[[1]]
 		}
-		lt = lt[[1]]
 	}
 
 	n = length(lt)
-	if(n == 1) {
-		stop_wrap("There should be at least two combination matrices.")
-	}
 
 	all_set_names = sort(unique(unlist(lapply(lt, set_name))))
 	if(any(sapply(lt, function(x) attr(x, "param")$set_on_rows) != attr(lt[[1]], "param")$set_on_rows)) {
 		stop_wrap("Combination sets should be all on rows or on columns.")
 	}
+	n_set = length(all_set_names)
 
 	lt = lapply(lt, function(m) {
 		if(attr(m, "param")$set_on_rows) {
@@ -1627,7 +1660,14 @@ normalize_comb_mat = function(...) {
 		}
 	})
 
-	all_comb_names = sort(unique(unlist(lapply(lt, comb_name))))
+	if(full_comb_sets) {
+		all_comb_names = full_comb_code(n_set)
+	} else {
+		all_comb_names = sort(unique(unlist(lapply(lt, comb_name))))
+	}
+	if(complement_set) {
+		all_comb_names = union(all_comb_names, strrep("0", n_set))
+	} 
 	lt = lapply(lt, function(m) {
 		if(attr(m, "param")$set_on_rows) {
 			m = m[, all_comb_names]
@@ -1636,7 +1676,11 @@ normalize_comb_mat = function(...) {
 		}
 		m[order.comb_mat(m)]
 	})
-	return(lt)
+	if(input_is_comb_mat) {
+		return(lt[[1]])
+	} else {
+		return(lt)
+	}
 }
 
 # == title
