@@ -16,6 +16,14 @@ This message can be suppressed by:
 ========================================
 ")	
 
+  	if("package:pheatmap" %in% search()) {
+  		msg = paste0(msg, 
+"! pheatmap() has been masked by ComplexHeatmap::pheatmap(). 90% of the arguments
+   in the original pheatmap() are identically supported in the new function. You 
+   can still use the original function by explicitly calling pheatmap::pheatmap().
+")
+  	}
+
     packageStartupMessage(msg)
 }
 
